/*
Copyright (c). The Regents of the
University of California (Regents). All Rights Reserved.

Permission to use, copy, modify, and distribute this software and its
documentation for educational, research, and not-for-profit purposes,
without fee and without a signed licensing agreement, is hereby granted,
provided that the above copyright notice, this paragraph and the
following two paragraphs appear in all copies, modifications, and
distributions. Contact The Office of Technology Licensing, UC Berkeley,
2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620, (510) 643-7201,
for commercial licensing opportunities. Created by Eric C. Anderson and
John A. Novembre,
Department of Integrative Biology, University of California, Berkeley.

IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT,
SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS,
ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
REGENTS HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY,
PROVIDED HEREUNDER IS PROVIDED "AS IS". REGENTS HAS NO OBLIGATION TO
PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*/



#define MAX_SEQUENCE_NAME_LENGTH 100
#define MAX_NEIGHBORS 40
#define MAX_NUM_MARKERS 1000


#define MISSING_GENOTYPE  	-1
#define AMBIG_HET			-2


// For the block from [a,b], this picks out the 0 or 1 which is the value
// of the SNP (missing data filled in) in the i-th set for that block at
// the k-th position in the block.
#define  SAME_SEQ_VAL(B,a,b,i,k)   B->Y[ B->S[b].Set[i][0] ] [ k+a ]


#define NUM_QPRIOR_DSNS 3
enum qtype_enum {           /* enums for the different "prior" dsns on q */
	MIXTURE,				/* for each block, choose the prior that minimizes the cost */
	STANDARD,     			/* the standard way of dealing with it---favors highly skewed haplotype freqs  */
	EWENS,	        /* the dsn that arises as a uniform dsn on ways of partitioning things up---favors
							   haplotype freqs that are more intermediate     */
	OCCUPANCY					/* the prior dsn is the multivariate Ewens dsn conditional on the number of types in the block */
};

typedef struct {
	int Idx;   // the idx of the canonical sequence of the sequences in this set
	int nC;   // the number of sequences compatible with the canonical sequence
	int *CoSeq;   // list of the indices of the compatible sequences
} compat_struct;

typedef struct {
		int N;   // the number canonical seqs
		compat_struct *CS; 
} canon_container;


typedef struct {
	int Idx;   // the index of the sequence that defines this set of canonical seqs
	int *Row;  // array of 0's and 1's.  if Row[i] is 1 then seq i is compatible with the canonical seq
	int RowSum;  // the sum of the row elements
} findcan_struct;


typedef struct  {
	int N;    // the number of canonical sequences at this block
	int *Nd;  // the number of sequences in the data set corresponding to each canonical sequence
	int **Idx;  // the indices of the sequences in the data set corresponding to each canonical sequence
} trans_info_struct;

typedef struct {
	int NumSets;
	int *NumInSet;
	int **Set;
} set_struct;

typedef struct {
	int Idx;
	double d;
} indexed_doubles_struct;

typedef struct bin_node{
	
	struct bin_node *Br[2];  // branches, labelled 0 and 1
	int Visits;  // for counting how many sequences end at a tip
	int Visits2;
	int Idx;     // for storing an index associated with the node.  May not be necessary
	
	int nOnes;
	int nZeroes; 
	
	int *SetPtr;
	
} bin_node;


typedef struct
{
	int NSWM;  // number of sequences with missing data
	int Init_NSWM;  // the number of sequences when the memory is allocated to the missing_data_info struct
	int *M_Idx;  // the indices of the sequences that have missing data
	int *NumMissingArray;   // array of numbers missing in each sequence which has at least one missing site.
	int **MissingIndices;  // for each sequence that has a missing site, we have an array of length NumMissing of the indices that
						   // are missing.
	int TotalMissing;
				
} missing_data_info;

typedef struct
{
	int NumTips;   // number of tips of the binary tree (basically the max possible number of archetypal sequences
	int NumSeq;   // The number of chromosomes in the data set
	int Length;   // the length of the block being considered
	int MaxNumTips;  // the number of rows that will be allocated to things
	
	int **TheTipSeqs;   // the sequences on the tips.  This has NumTips rows and length cols, but when it gets
						// allocated, it will have NumTips rows and NumMarkers cols, so it never has to get
						// reallocated again.
	int **Diffs;    // array of number of mismatches between each TipSeq and each chromosome.  Has
					// NumTips rows and NumSeq cols.  
	
	int NumA;       // the number of seqs in the set of archetypes
	int *ASet;  	// a group of integers telling which of TheTipSeqs are included
					// in the set we will call "archetypal"
	int *MinMs;  // the number of mismatches given the archetypes in ASet.  subscripted
				   // for different sequences in the data set
	int SumMinMs;  // the sum across the different sequences of MinMs
	
	
	
	
	int PNumA;	// for proposed versions of the ASet
	int *PASet;    
	int *PMinMs;
	int PSumMinMs;
	 
} arch_struct;


typedef struct
{
	int NumMarkers;	// the number of markers	
	int NumSeq;   	// the number of "sequences"
	
	int **Yobs;  // the observed values of the SNP's.  These will be 0,1 for SNP's, and something
					// less than 0 (like -1 or -2) if missing. 
					
	char **Names;  // to store the names of each sequence
	
	int **Y;	 // to hold the data after the missing values have been filled int
	
	double **E;  // the matrix of entropies that have been computed.
	double **CanonSeqMultParCost;
	enum qtype_enum **ChosenQPrior;
	double **SumOfCodeLengths;
	
	double **CodeLengths;  // an array of Code Lengths for 0's and 1's at each marker, computed from the
						   // empirical frequency of the 0's and 1's.   
	  
	int **NT;	 // the matrix of the number of types seen. 
	
	int NumBounds;  // the number of boundaries inferred		
	int *Bounds;	// the array of boundaries inferred (in sorted order).	
	
	
	bin_node *Nodes;
	bin_node **Tips;
	bin_node ***TipsArray;
	
	int *TipsNums;
	int *TipsNums2;
	
	int **TipSeqs;	
	
	missing_data_info *AH;  // this is the "missing due to ambiguous heterozygosity pile
	missing_data_info *TM;  // this is for the data that are "truly missing"
	
	int AH_WindowWidth;
	int TM_WindowWidth;
	
	arch_struct *A;
	
	set_struct *S;
	
	int **Sets;
	
	canon_container *Canon;
	
	findcan_struct **FC;
	findcan_struct **FCP;  /*  the "previous" FC, used for determining transition rates */
	int FCP_filled;
	
	double EntropyWeightFactor;
	
	trans_info_struct **Tr;

	char *InfileName;
	 
} MDL_struct;





// FUNCTION PROTOTYPES

/******************************************
MEMORY MANAGEMENT, ALLOCATION, DE-ALLOCATION
******************************************/
MDL_struct *AllocMDL_struct(int NumSeq, int NumMarkers);
void DeAllocMissingDataInfo(missing_data_info *M);
double **AllocEntMat(int NumMarkers);
int **AllocNTMat(int NumMarkers);


/******************************************
VARIABLE INITIALIZATION
******************************************/
void InitEntMat(double **mat, int NumMarkers);
void InitNTMat(int **NT, int NumMarkers);
void InitNode(struct bin_node *N);

/******************************************
READING DATA IN
******************************************/
MDL_struct *ReadSeqsAsBits(char *FileName);



/******************************************************
IMPLEMENTING THE MINIMUM DESCRIPTION LENGTH METHOD
******************************************************/
void GreedyFindByMDL(MDL_struct *B);
double ModelBitsAndDataBits(int n, int *A, MDL_struct *B);
void ComputeBlockEntropyAndNT_WithSeqs(int a, int b, MDL_struct *B);

/*********************************
COMPUTING CODE LENGTHS
*********************************/
void ComputeCodeLengths(MDL_struct *B, int **Data);

/***********************************
	SORTING THE BLOCK TYPES, ETC
***********************************/
int CompareBinNodesForVisitSort(const void *vT1, const void *vT2);


/******************************************
FOR IMPLEMENTING THE DP ALGORITHM
******************************************/
// this fills in all the matrices that need to be filled
// in to implement the DP algorithm.  
void FillInMatrices(MDL_struct *B);



/*********************************************
DEALING WITH MISSING DATA, in file MDL_Blocks_MissingData.c
*********************************************/
// a function to look at the rates of missingness at each SNP separately
int *NumMissingByColumn(MDL_struct *B, int TypeOfMissing);

// Keeping track of which ones are missing:
missing_data_info *CollectMissingSites(MDL_struct *B, int TypeOfMissing);

// basic operations for determining probabilities of 0 or 1 in the missing site:
void DesignateNeighbors(int I, int M, int NN, int *Neighbors, int n, MDL_struct *B);
void ComputeProbsForHoleY(int I, int M, int NN, int *Neighbors, MDL_struct *B, 
						double ZeroInit, double OneInit, double *ZeroProbs, double *OneProbs);
						
// functions for drawing 0's or 1's to fill the holes due to unresolved heterozygosity
int DrawFromMostCertainAmbigH(double *ZP1, double *OP1, double *ZP2, 
		double *OP2, int NN);
void FillAmbigHetHole(MDL_struct *B, int V, int J, int NN);
int FillRandomAmbigHet(MDL_struct *B, int NN);

// functions for drawing 0's or 1's to fill the holes due to truly missing data
int DrawFromMostCertainSingle(double *ZeroProbs, double *OneProbs, int NN);
void FillTrulyMissingHole(MDL_struct *B, int V, int J, int NN);
int FillRandomTrulyMissing(MDL_struct *B, int NN);

// operations on arrays to help trim them down as we fill the missing holes
void DeleteIntElement(int *A, int x, int n);
void FreeAndDeleteIntStarElement(int **A, int x, int n);

// a single function that fills all the missing data then deals with de-allocating
void FillAllMissingHoles(MDL_struct *B);




/*******************************************************
DOING THE ARCH-DIFF METHOD  in arch_stuff.c
******************************************************/
arch_struct *AllocArchStruct(int NumSeq, int NumTips, int NumMarkers);
void FreeArchStruct(arch_struct *A);
void TransferToArch(arch_struct *A, MDL_struct *B, int NumTips, int a, int b);
void PrintArchStruct(arch_struct *A, char *FileName, int a, int b);

/********************************************
MORE EFFICIENT VERSIONS FOR DP METHOD
********************************************/
int CompIndexedDoubles(const void *A, const void *B);
int CompIndexedDoublesAscendng(const void *A, const void *B);
double CostWithMixture(int z, int a, int b, MDL_struct *B, double *DataCostOutput, double *DeltaCost, 
			double *PCost, int *NumDelts, double *log2NabSum, double *FixedBitCost, double *KLD);
void StoreTransInfo(findcan_struct **F, int N, trans_info_struct *T, int NumSeq);
void MoreEfficientEntropy(int a, MDL_struct *B, int btop, int FillinMats);
void MoreEfficientFillInMatrices(MDL_struct *B);
void StandardScoreFromSetsArray(int a, MDL_struct *B);
void CompileCompatibleSeqs(int a, MDL_struct *B);

void InitFindCan(int N, int b, findcan_struct **FC, MDL_struct *B);
void FreeCompatSwitches(int **temp, int NumSeq);
int **AllocAndInitCompatSwitches(int NumSeq);
int IsCompatible(int i, int a, int b, int *CS, int Idx, MDL_struct *B);
void ComputeEntropyByCompats(int a, MDL_struct *B);
int CompareFindCanRows(const void *vT1, const void *vT2);
int SortAndChooseCanonicals(findcan_struct **FC, MDL_struct *B, int N);
double EntropyOfSortedCanonicals(findcan_struct **FC, int N);
double SumCodeLengthOfSortedCanonicals(findcan_struct **FC, int N, int a, int b, MDL_struct *B);
void ReportCanonicalSeqs(int a, MDL_struct *B, int b, char *FileName);
void IncrementMismatchMatrix(int N, findcan_struct **FC, int **Mis, int a, int b, MDL_struct *B);
void ComputeCodeLengthsUsingDiffs(int NC, findcan_struct **FC, int **Mis, int a, int b, MDL_struct *B);
double ComputeCanonSeqMultParCost(findcan_struct **FC, int N, int S, enum qtype_enum *Q);


void PrintCostsLatex(MDL_struct *B, int *Bounds, int NumBounds, char *FileName);
char *qtype2str(enum qtype_enum Q);

int PassesTheFourGameteTest(MDL_struct *B, int a, int b);
void DoWang(MDL_struct *B);
void PrintEasyGrepLine(MDL_struct *B, long seed1, long seed2);
int SeqPairsPassFGT(MDL_struct *B,int a,int b);

int NoMissingData(MDL_struct *B, int theSeq, int a, int b);


// global variables
#ifdef UN_EXTERN
	#define GLOB
#else
	#define GLOB extern
#endif

GLOB int gVerboseFlag;  // 0 to have the program not speak too much.  1 to have it print
					   // out information as it is progressing.
GLOB int gCloserLook;
GLOB FILE *gCloserLookFile; 
GLOB char gCloserLookFileName[100]; 

GLOB int gPrintTestBounds;
GLOB FILE *gPrintTestBoundsFile;
GLOB int *gDP_Inferred_Bounds; 
GLOB int gDP_NumBoundsInferred;
GLOB char gCanonSeqFileName[100];
GLOB double gModelBooster;  // factor that gets multiplied to the model cost
GLOB double gDP_BitScore;

GLOB int gGreedy; // 1/0, denotes whether to use greedy algorithm or not

GLOB double *gJackknifeScores;

GLOB char **gDalyBases;  // wacky thing to be able to print out the inferred canonical
						 // seqs of daly as A, C, G, and T's
GLOB enum qtype_enum gQPrior;

GLOB double *gKLDs;   /* Kullback-Leibler Divergences  */
GLOB long gSeed1, gSeed2;
GLOB int gPrevReportedA;  /* used in ReportCanonicalSeqs */

int ShiftQPrior(int a);

/******************************************
For resolving cycles in the iteration over gR
******************************************/
GLOB int gR;
void ResolveCycle(int *Rvisited, double *Rscores, MDL_struct *B);

double pwise_d_prime(MDL_struct *B, int a, int b, int *ssize);

/******************************************
For optimizing description length
******************************************/
void DP_Blocks(MDL_struct *S);
void GreedyBlocks(MDL_struct *S);

/* legacy */
void SimAnnBlocks(MDL_struct *S,double initT, int Imax,int Nk);

/******************************************
For assessing confidence in boundary calls
******************************************/
void JackBounds(MDL_struct *S);
double ScoreBinaryBounds(int* BinBounds,MDL_struct *S);
double ScoreBounds(int* Bounds,int nBounds,MDL_struct *S);

/******************************************
For handling command-line parameters
******************************************/
void usage();



/******************************************
For trying some k-means clustering 
******************************************/
void cluster_a2b(MDL_struct *M, int NC, int a, int b);
GLOB int gUseKMeansClustering;

/**********************************
For doing Wang's method with a threshold amount of
incompatibility required 
***********************************/
int CountIncompatibilities(MDL_struct *B, int a, int b, int *NumNoMissing);
void DoWangWithSlop(MDL_struct *B, double slop);

/****************** 
to compute heterozygosities at the different sites  
******************/
void PrintHeterozygosities(MDL_struct *B);

